/*
 *	heap.c
 */


#include <stdio.h>
#include "scheme.h"
#include "arch.h"
#include "heap.h"
#include "reg.h"			/* XXX for reg_ stuff */
#include "memory.h"			/* for create_memory declaration */

#include "object.h"			/* XXX just for make_stob , QUIE_OBJ */
#include "data.h"			/* XXX just for make_stob */
#include "vmio.h"


int newspace_begin;			/* beginning of current consing area */
int newspace_end;			/* end of current consing area */
int oldspace_begin;			/* beginning of last consing area */
int oldspace_end;			/* end of last consing area */

static FILE *fp;
static int new_heap;
static old_bits_per_byte;
static old_endedness;


void
initialize_heap(size)
int size;
{
	int halfsize;
	extern void initialize_io_system();

	memory = initialize_memory(size, QUIESCENT_OBJ);

	initialize_io_system();

	halfsize = (memory->end - memory->begin) / 2;
	newspace_begin = memory->begin;
	newspace_end = memory->begin + halfsize;
	oldspace_begin = newspace_end;
	oldspace_end = oldspace_begin + halfsize;

	reg_heap = newspace_begin;
	reg_limit = newspace_end - MARGIN;

	reg_root = QUIESCENT_OBJ;
	reg_stack = QUIESCENT_OBJ;
}


/*** inline ***/
INDEX_AFTER_HEADER(stob)
int stob;
{
	assert(STOB_P(stob));
	return DESCRIPTOR_DATA(stob);
}


swap(data)
unsigned int data;
{
	unsigned char *p, temp;

	p = (unsigned char *) &data;
	temp = p[0]; p[0] = p[3]; p[3] = temp;
	temp = p[1]; p[1] = p[2]; p[2] = temp;
	return data;
}


adjust(desc, index_delta)
int desc;
int index_delta;
{
	if (! STOB_P(desc))
		return desc;

	return MAKE_STOB_DESCRIPTOR(INDEX_AFTER_HEADER(desc) + index_delta);
}


read_heap(filename)
char *filename;
{
	int old_level, old_bytes_per_cell;
	int old_begin, old_heap, old_root, old_stack;
	int delta, new_limit;
	int desc;

	if ((fp = fopen(filename, "r")) == (FILE *) NULL)
	{
		printf("file = %s\n", filename);
		error("Could not open '%s' to read image.", filename);
	}

	if (fscanf(fp, "%d\n", &old_endedness) != 1)
		error("reading endedness information.");

	if (fscanf(fp, "%d\n", &old_level) != 1)
		error("reading level information");
	if (fscanf(fp, "%d\n", &old_bits_per_byte) != 1)
		error("reading bits per byte information");
	if (fscanf(fp, "%d\n", &old_bytes_per_cell) != 1)
		error("reading bytes per cell information");
	if (fscanf(fp, "%d\n", &old_begin) != 1)
		error("reading beginning of heap information");
	if (fscanf(fp, "%d\n", &old_heap) != 1)
		error("reading heap pointer information");
	if (fscanf(fp, "%d\n", &old_root) != 1)
		error("reading rooting information");
	if (fscanf(fp, "%d\n", &old_stack) != 1)
		error("reading stack information");

	if (old_level != DUMP_LEVEL)
		error("Format of image is incompatible with this system.");

	if (old_bytes_per_cell != BYTES_PER_CELL)
		error("Incompatible bytes per cell");

	initialize_heap(HEAP_SIZE);

	delta = newspace_begin - old_begin;
	new_heap = old_heap + delta;
	new_limit = newspace_end - MARGIN;

	if (new_heap >= new_limit)
		error("Heap not big enough to restore this image.");

	reg_root = adjust(old_root, delta);
	reg_stack = adjust(old_stack, delta);

	while (reg_heap < new_heap)
	{
		if (fscanf(fp, "%d\n", &desc) != 1)
			error("Could not read descriptor. Heap ptr = %d",
			      reg_heap);

		desc = adjust(desc, delta);

		memory->data[reg_heap++] = desc;

		if (B_VECTOR_HEADER_P(desc))
			read_byte_vector(desc);
	}
	putchar('\n');
}


read_byte_vector(desc)
int desc;
{
	int end, ndesc;

	end = reg_heap + HEADER_AUNITS(desc);
	while (reg_heap < end && reg_heap < new_heap)
	{
		if (fscanf(fp, "%d\n", &ndesc) != 1)
			error("Could not read descriptor.");

		if (old_bits_per_byte == BITS_PER_BYTE
		    && old_endedness == ENDEDNESS)
			memory->data[reg_heap++] = ndesc;
		else
			memory->data[reg_heap++] = adjust_bytes(ndesc);
	}
}


adjust_bytes(desc)
int desc;
{
	unsigned int old_byte_mask, value;

	old_byte_mask = ~ (0xffffffff << old_bits_per_byte);

	value = (desc >> (3 * old_bits_per_byte))
		& old_byte_mask;
	value <<= BITS_PER_BYTE;
	value |= (desc >> (2 * old_bits_per_byte))
		& old_byte_mask;
	value <<= BITS_PER_BYTE;
	value |= (desc >> old_bits_per_byte)
		& old_byte_mask;
	value <<= BITS_PER_BYTE;
	value |= desc & old_byte_mask;

	if (old_endedness == ENDEDNESS)
		return value;
	else
		return swap(value);
}



write_heap(filename)
char *filename;
{
	int i;
	int p, desc, z;
	FILE *fp;

	if ((fp = fopen(filename, "w")) == (FILE *) NULL)
		error("Could not open '%s' to write image", filename);

	for (i = 2; i < MAX_PORTS; i++)	/* XXX skip input and output */
		if (port_table[i].fp)
			close_port_noisily(port_table[i].port);

	fprintf(fp, "%d\n", ENDEDNESS);

	fprintf(fp, "%d\n", DUMP_LEVEL);
	fprintf(fp, "%d\n", BITS_PER_BYTE);
	fprintf(fp, "%d\n", BYTES_PER_CELL);
	fprintf(fp, "%d\n", newspace_begin);
	fprintf(fp, "%d\n", reg_heap);

	fprintf(fp, "%d\n",
		MAKE_DESCRIPTOR(TAG_STOB,
				(CELL *) address_after_header(reg_root)
				- &memory->data[0]));
	fprintf(fp, "%d\n",
		MAKE_DESCRIPTOR(TAG_STOB,
				(CELL *) address_after_header(reg_stack)
				- &memory->data[0]));


	for (p = newspace_begin; p < reg_heap; )
	{
		desc = memory->data[p++];

		if (! STOB_P(desc))
			fprintf(fp, "%d\n", desc);
		else
			fprintf(fp, "%d\n", 
				MAKE_DESCRIPTOR(TAG_STOB,
					(CELL *) address_after_header(desc)
					- &memory->data[0]));

		if (B_VECTOR_HEADER_P(desc))
			for (z = p + HEADER_AUNITS(desc); p < z; p++)
				fprintf(fp, "%d\n", memory->data[p]);
	}
	(void) fclose(fp);
}
