/*
 *	interp.c
 */


#include <setjmp.h>
#include <signal.h>
#include "scheme.h"
#include "op.h"				/* primitive machine operations */
#include "ex.h"				/* exceptions */
#include "arch.h"			/* addressing architecture */
#include "reg.h"			/* machine registers */
#include "memory.h"
#include "object.h"
#include "data.h"
#include "stack.h"
#include "struct.h"
#include "heap.h"			/* available_p */
#include "vmio.h"			/* for FOR_INPUT FOR_OUTPUT */


/*
 *  The following is for inline expansion of procedures with the
 *  GNU C compiler.  In order for procedures to be expanded in place
 *  they must be defined as inlineable prior to their use, hence we
 *  include C files here.  Yuck.
 */
#ifdef INLINE
#include "object.c"
#include "data.c"
#include "stack.c"
#include "struct.c"
#endif


#define BITS_USED_PER_BYTE	7

#define TEMPLATE_CODE(tem)		VECTOR_REF((tem), 0)
#define TEMPLATE_NAME(tem)		VECTOR_REF((tem), 1)

#define CONTINUATION_SIZE		STACK_VECTOR_SIZE(5)

#define CONTINUATION_PC(c)		VECTOR_REF((c), 0)
#define CONTINUATION_TEMPLATE(c)	VECTOR_REF((c), 1)
#define CONTINUATION_ENV(c)		VECTOR_REF((c), 2)
#define CONTINUATION_CONT(c)		VECTOR_REF((c), 3)

#define ISTATE_SIZE	(PUSH_SIZE + CONTINUATION_SIZE + STACK_VECTOR_SIZE(6))

#define ISTATE_CONT(i)			VECTOR_REF((i), 0)
#define ISTATE_NARGS(i)			VECTOR_REF((i), 1)
#define ISTATE_VAL(i)			VECTOR_REF((i), 2)
#define ISTATE_ARG2(i)			VECTOR_REF((i), 3)
#define ISTATE_ARG3(i)			VECTOR_REF((i), 4)
#define ISTATE_EI(i)			VECTOR_REF((i), 5)
#define ISTATE_DS(i)			VECTOR_REF((i), 6)

#define GET_EXCEPTION_HANDLERS()	VECTOR_REF(reg_root, 1)
#define GET_INTERRUPT_HANDLERS()	VECTOR_REF(reg_root, 2)


#define REG_PC		(program_counter -				\
			 (BYTE *) address_after_header(reg_code))


extern int verbose_flag;

jmp_buf interpreter_context;

BYTE *program_counter;

void handle_interrupt();
void raise_exception();


static void
push_continuation(pc)
int pc;
{
	reg_cont = PUSH_VECTOR_4(enter_fixnum(pc),
				 reg_template,
				 reg_env,
				 reg_cont);
}


static void
pop_continuation()
{
	int cont;

	reg_stack = reg_cont;
	cont = POP_VECTOR(4);

	reg_template = CONTINUATION_TEMPLATE(cont);
	reg_code = TEMPLATE_CODE(reg_template);
	program_counter = 
		((BYTE *) address_after_header(reg_code))
			+ extract_fixnum(CONTINUATION_PC(cont));

	reg_env = CONTINUATION_ENV(cont);
	reg_cont = CONTINUATION_CONT(cont);
}



static CELL
push_istate()
{
	push_continuation(REG_PC);
	return PUSH_VECTOR_7(reg_cont,
			     enter_fixnum(reg_nargs),
			     reg_val,
			     reg_arg2,
			     reg_arg3,
			     enter_fixnum(reg_enabled_interrupts),
			     reg_dynamic_state);
}


static void
pop_istate()
{
	int istate;

	istate = POP_VECTOR(7);		/* magic num??? */
	reg_cont = ISTATE_CONT(istate);
	reg_nargs = ISTATE_NARGS(istate);
	reg_enabled_interrupts = ISTATE_EI(istate);
	reg_val = ISTATE_VAL(istate);
	reg_arg2 = ISTATE_ARG2(istate);
	reg_arg3 = ISTATE_ARG3(istate);
	reg_dynamic_state = ISTATE_DS(istate);
	pop_continuation();
}


static void
clear_registers()
{
	reg_cont = UNSPECIFIED_OBJ;
	reg_nargs = UNSPECIFIED_OBJ;
	reg_val = UNSPECIFIED_OBJ;
	reg_arg2 = UNSPECIFIED_OBJ;
	reg_arg3 = UNSPECIFIED_OBJ;
	reg_enabled_interrupts = 0;
	reg_template = UNSPECIFIED_OBJ;
	reg_code = UNSPECIFIED_OBJ;
	program_counter = (unsigned char *) NULL;
	reg_env = UNSPECIFIED_OBJ;
	reg_cont = UNSPECIFIED_OBJ;
	reg_pending_interrupt = INT_NONE;
}


#define THIS_BYTE()		((int) *program_counter)
#define NEXT_BYTE()		((int) *program_counter++)
#define PREVIOUS_BYTE()		(program_counter--)
#define NEXT_LITERAL()		VECTOR_REF(reg_template, NEXT_BYTE())


static int
next_offset()
{
	int high;

	high = NEXT_BYTE();
	return (high << BITS_USED_PER_BYTE) | NEXT_BYTE();
}


/*
 *  Environment access.
 */

#define MAKE_RIB		MAKE_VECTOR
#define RIB_REF			VECTOR_REF
#define RIB_SET			VECTOR_SET
#define RIB_PARENT(rib)		RIB_REF(rib, 0)


static CELL
env_back(env, back)
int env, back;
{
	int i;
	int temp = env;

	for (i = back; i > 0; i--)
		temp = RIB_PARENT(temp);
	return temp;
}


static void
ensure_space(space)
{
	if (! AVAILABLE_P(space))
	{
		allow_marginal_allocation();
		(void) push_istate();
		collect();
		if (! AVAILABLE_P(space))
			raise_exception(EX_HEAP_OVERFLOW);
		pop_istate();
	}
}


static void
ensure_stack_space(space)
{
	if (! AVAILABLE_ON_STACK_P(space + ISTATE_SIZE))
	{
		allow_marginal_allocation();
		(void) push_istate();
		collect();
		if (! AVAILABLE_ON_STACK_P(space + ISTATE_SIZE))
			raise_exception(EX_HEAP_OVERFLOW);
		pop_istate();
	}
}



static
register_interrupt()
{
	extern int waiting_in_read;
	extern jmp_buf read_char_context;

	reg_pending_interrupt = INT_KEYBOARD;

	if (waiting_in_read)
		longjmp(read_char_context, 1);
}


static void
perform_application()
{
	if (! CLOSURE_P(reg_val))
		raise_exception(EX_BAD_PROCEDURE);
	reg_env = CLOSURE_ENV(reg_val);
	reg_template = CLOSURE_TEMPLATE(reg_val);
	reg_code = TEMPLATE_CODE(reg_template);
	program_counter = (BYTE *) address_after_header(reg_code);
}



void
interpret()
{
	clear_registers();
	pop_istate();
	if (! AVAILABLE_ON_STACK_P(ISTATE_SIZE))
		error("Out of memory before interpreting bytecodes.");

	(void) signal(SIGINT, register_interrupt);
	(void) setjmp(interpreter_context);

    again:


#ifdef notdef
	int opcode;
	opcode = NEXT_BYTE();
	if (verbose_flag)
		print_info(opcode);
#endif

	switch(NEXT_BYTE())
	{
	    case OP_CHECK_NARGS_EQ:
		if (reg_nargs != NEXT_BYTE())
			raise_exception(EX_WRONG_NUMBER_OF_ARGUMENTS);
		break;
		
	    case OP_CHECK_NARGS_GE:
		if (reg_nargs < NEXT_BYTE())
			raise_exception(EX_WRONG_NUMBER_OF_ARGUMENTS);
		break;
		
	    case OP_MAKE_ENV:
		{
			int i, rib;

			reg_nargs = THIS_BYTE();
			ensure_space(VECTOR_SIZE(reg_nargs + 1));
			NEXT_BYTE();
			rib = MAKE_RIB(reg_nargs + 1);
			RIB_SET(rib, 0, reg_env);
			reg_env = rib;
			for (i = reg_nargs; i >= 1; i--)
				RIB_SET(rib, i, POP());
		}
		break;
		
	    case OP_MAKE_REST_LIST:
		{
			int min_nargs, l, i;

			min_nargs = THIS_BYTE();
			ensure_space(PAIR_SIZE * (reg_nargs - min_nargs));
			NEXT_BYTE();
			l = NULL_OBJ;
			for (i = reg_nargs; i > min_nargs; i--)
				l = cons(POP(), l);
			PUSH(l);	/* kludge */
			reg_nargs = min_nargs + 1;
		}
		break;
		
	    case OP_LITERAL:
		reg_val = NEXT_LITERAL();
		break;
		
	    case OP_LOCAL:
		{
			int back;

			back = NEXT_BYTE();
			reg_val = RIB_REF(env_back(reg_env, back),
					  NEXT_BYTE());
			if (reg_val == UNASSIGNED_MARKER_OBJ)
				raise_exception(EX_UNASSIGNED_LOCAL);
		}
		break;
		
	    case OP_SET_LOCAL:
		{
			int back;

			back = NEXT_BYTE();
			RIB_SET(env_back(reg_env, back),
				NEXT_BYTE(),
				reg_val);
			reg_val = UNSPECIFIED_OBJ;
		}
		break;
		
	    case OP_GLOBAL:
		{
			int cell;

			cell = NEXT_LITERAL();
			reg_val = CELL_CONTENTS(cell);
			if (UNDEFINED_P(reg_val)) /* unbound or unassigned */
			{
				reg_arg2 = cell;
				if (reg_val == UNASSIGNED_MARKER_OBJ)
					raise_exception(EX_UNASSIGNED_GLOBAL);
				else
					raise_exception(EX_UNBOUND_GLOBAL);
			}
		}
		break;
		
	    case OP_SET_GLOBAL:
		{
			int cell;

			cell = NEXT_LITERAL();
			if (CELL_CONTENTS(cell) == UNBOUND_MARKER_OBJ)
			{
				reg_arg2 = cell;
				raise_exception(EX_UNBOUND_GLOBAL);
			}
			SET_CONTENTS(cell, reg_val);
			reg_val = UNSPECIFIED_OBJ;
		}
		break;
		
	    case OP_CLOSURE:
		ensure_space(CLOSURE_SIZE);
		reg_val = make_closure(NEXT_LITERAL(), reg_env);
		break;

	    case OP_PUSH:
		ensure_stack_space(PUSH_SIZE);
		PUSH(reg_val);
		break;
		
	    case OP_POP:
		reg_val = POP();
		break;
		
	    case OP_MAKE_CONT:
		ensure_stack_space(CONTINUATION_SIZE);
		push_continuation(REG_PC + next_offset());
		break;

	    case OP_CALL:
		if (reg_pending_interrupt != INT_NONE)
			handle_interrupt();
		reg_nargs = THIS_BYTE();
		perform_application();
		break;
		
	    case OP_JUMP_IF_FALSE:
		{
			int offset;

			offset = next_offset();
			if (FALSE_P(reg_val))
				program_counter += offset;
		}
		break;

	    case OP_JUMP:
		program_counter += next_offset();
		break;

	    case OP_RETURN:
		pop_continuation();
		break;
		
	    case OP_PUSH_CONT:
		ensure_stack_space(PUSH_SIZE);
		PUSH(reg_cont);
		break;
		
	    case OP_POP_CONT:
		reg_cont = POP();
		break;
		
	    case OP_SPREAD_ARGS:
		{
			int length, list;
			
			reg_nargs = NEXT_BYTE();
			list = reg_val;
			for (length = 0; ! (list == NULL_OBJ); length++)
				list = CDR(list);
			ensure_stack_space(PAIR_SIZE * length);
			while (PAIR_P(reg_val))
			{
				PUSH(CAR(reg_val));
				reg_val = CDR(reg_val);
				reg_nargs++;
			}
			PUSH(enter_fixnum(reg_nargs));
		}
		break;
		
	    case OP_N_CALL:
		if (reg_pending_interrupt != INT_NONE)
			handle_interrupt();
		reg_nargs = extract_fixnum(POP());
		perform_application();
		break;
		
	    case OP_NATIVE:
		error("Native code encountered.  (currently unsupported)");
		break;

		/*
		 *  Scalar primitives.
		 */

	    case OP_EQ_P:
		reg_arg2 = POP();
		reg_val = ENTER_BOOLEAN(reg_val == reg_arg2);
		break;

	    case OP_INTEGER_P:
		reg_val = ENTER_BOOLEAN(FIXNUM_P(reg_val));
		break;

	    case OP_NUMBER_P:
		reg_val = ENTER_BOOLEAN(NUMBER_P(reg_val));
		break;

	    case OP_EQUAL:
		reg_arg2 = POP();
		if (! NUMBER_P(reg_val) || ! NUMBER_P(reg_arg2))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = ENTER_BOOLEAN(reg_val == reg_arg2);
		break;

	    case OP_LESS_THAN:
		reg_arg2 = POP();
		if (! NUMBER_P(reg_arg2) || ! NUMBER_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = ENTER_BOOLEAN(reg_arg2 < reg_val);
		break;

	    case OP_PLUS:
		{
			int temp;

			reg_arg2 = POP();
			if (! NUMBER_P(reg_val) || ! NUMBER_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			temp = extract_fixnum(reg_arg2) 
				+ extract_fixnum(reg_val);
			if (OVERFLOWS_P(temp))
				raise_exception(EX_ARITHMETIC_OVERFLOW);
			else
				reg_val = enter_fixnum(temp);
		}
		break;

	    case OP_MINUS:
		{
			int temp;

			reg_arg2 = POP();
			if (! NUMBER_P(reg_val) || ! NUMBER_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			temp = extract_fixnum(reg_arg2)
				- extract_fixnum(reg_val);
			if (OVERFLOWS_P(temp))
				raise_exception(EX_ARITHMETIC_OVERFLOW);
			else
				reg_val = enter_fixnum(temp);
		}
		break;

	    case OP_QUOTIENT:
		{
			int temp;

			reg_arg2 = POP();
			if (! INTEGER_P(reg_val) || ! INTEGER_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			temp = extract_fixnum(reg_arg2)
				/ extract_fixnum(reg_val);
			if (OVERFLOWS_P(temp))
				raise_exception(EX_ARITHMETIC_OVERFLOW);
			else
				reg_val = enter_fixnum(temp);
		}
		break;

	    case OP_REMAINDER:
		{
			int temp;

			reg_arg2 = POP();
			if (! NUMBER_P(reg_val) || ! NUMBER_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			temp = extract_fixnum(reg_arg2)
				% extract_fixnum(reg_val);
			if (OVERFLOWS_P(temp))
				raise_exception(EX_ARITHMETIC_OVERFLOW);
			else
				reg_val = enter_fixnum(temp);
		}
		break;

	    case OP_TIMES:
		{
			int temp;

			reg_arg2 = POP();
			if (! NUMBER_P(reg_val) || ! NUMBER_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);

			temp = times_carefully(reg_val, reg_arg2);
			if (temp == ILLEGAL_FIXNUM)
				raise_exception(EX_ARITHMETIC_OVERFLOW);
			reg_val = temp;
		}
		break;

	    case OP_CHAR_P:
		reg_val = ENTER_BOOLEAN(CHAR_P(reg_val));
		break;

	    case OP_CHAR_EQUAL_P:
		reg_arg2 = POP();
		if (! CHAR_P(reg_arg2) || ! CHAR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = ENTER_BOOLEAN(extract_char(reg_arg2)
					== extract_char(reg_val));
		break;

	    case OP_CHAR_LESS_P:
		reg_arg2 = POP();
		if (! CHAR_P(reg_arg2) || ! CHAR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = ENTER_BOOLEAN(extract_char(reg_arg2)
					< extract_char(reg_val));
		break;

	    case OP_CHAR_TO_ASCII:
		if (! CHAR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = enter_fixnum(extract_char(reg_val));
		break;

	    case OP_ASCII_TO_CHAR:
		if (! FIXNUM_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = ENTER_CHAR(extract_fixnum(reg_val));
		break;
		
	    case OP_EOF_OBJECT_P:
		reg_val = ENTER_BOOLEAN(reg_val == EOF_OBJ);
		break;

		/*
		 *  Data manipulation.
		 */

	    case OP_PAIR_P:
		reg_val = ENTER_BOOLEAN(PAIR_P(reg_val));
		break;

	    case OP_CONS:
		reg_arg2 = POP();
		ensure_space(PAIR_SIZE);
		reg_val = cons(reg_arg2, reg_val);
		break;

	    case OP_CAR:
		if (! PAIR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = CAR(reg_val);
		break;

	    case OP_CDR:
		if (! PAIR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = CDR(reg_val);
		break;

	    case OP_SET_CAR:
		reg_arg2 = POP();
		if (! PAIR_P(reg_arg2))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		SET_CAR(reg_arg2, reg_val);
		reg_val = UNSPECIFIED_OBJ;
		break;

	    case OP_SET_CDR:
		reg_arg2 = POP();
		if (! PAIR_P(reg_arg2))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		SET_CDR(reg_arg2, reg_val);
		reg_val = UNSPECIFIED_OBJ;
		break;

	    case OP_SYMBOL_P:
		reg_val = ENTER_BOOLEAN(SYMBOL_P(reg_val));
		break;

	    case OP_MAKE_SYMBOL:
		ensure_space(SYMBOL_SIZE);
		reg_val = make_symbol(reg_val);
		break;

	    case OP_SYMBOL_TO_STRING:
		if (! SYMBOL_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = SYMBOL_TO_STRING(reg_val);
		break;

	    case OP_CELL_P:
		reg_val = ENTER_BOOLEAN(CELL_P(reg_val));
		break;

	    case OP_MAKE_CELL:
		reg_arg2 = POP();
		ensure_space(CELL_SIZE);
		reg_val = make_cell(reg_arg2, reg_val);
		break;

	    case OP_CELL_NAME:
		if (! CELL_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = CELL_NAME(reg_val);
		break;

	    case OP_CONTENTS:
		if (! CELL_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = CELL_CONTENTS(reg_val);
		break;

	    case OP_SET_CONTENTS:
		reg_arg2 = POP();
		if (! CELL_P(reg_arg2))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		SET_CONTENTS(reg_arg2, reg_val);
		reg_val = UNSPECIFIED_OBJ;
		break;

	    case OP_CLOSURE_P:
		reg_val = ENTER_BOOLEAN(CLOSURE_P(reg_val));
		break;

	    case OP_MAKE_CLOSURE:
		reg_arg2 = POP();
		ensure_space(CLOSURE_SIZE);
		reg_val = make_closure(reg_arg2, reg_val);
		break;

	    case OP_CLOSURE_ENV:
		if (! CLOSURE_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = CLOSURE_ENV(reg_val);
		break;

	    case OP_CLOSURE_TEMPLATE:
		if (! CLOSURE_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = CLOSURE_TEMPLATE(reg_val);
		break;

	    case OP_CODE_VECTOR_P:
		reg_val = ENTER_BOOLEAN(CODE_VECTOR_P(reg_val));
		break;

	    case OP_MAKE_CODE_VECTOR:
		{
			int i, len, vec, value;

			reg_arg2 = POP();
			if (! FIXNUM_P(reg_arg2)
			    || (len = extract_fixnum(reg_arg2)) < 0)
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			if (! FIXNUM_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			ensure_space(CODE_VECTOR_SIZE(len));
			vec = MAKE_CODE_VECTOR(len);
			value = extract_fixnum(reg_val);
			for (i = 0; i < len; i++)
				CODE_VECTOR_SET(vec, i, value);
			reg_val = vec;
		}
		break;

	    case OP_CODE_VECTOR_LENGTH:
		if (! CODE_VECTOR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = enter_fixnum(CODE_VECTOR_LENGTH(reg_val));
		break;

	    case OP_CODE_VECTOR_REF:
		{
			int index;

			reg_arg2 = POP();
			if (! CODE_VECTOR_P(reg_arg2) || ! FIXNUM_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			index = extract_fixnum(reg_val);
			if (! VALID_INDEX_P(index,
					    CODE_VECTOR_LENGTH(reg_arg2)))
				raise_exception(EX_INDEX_OUT_OF_RANGE);
			reg_val = enter_fixnum(CODE_VECTOR_REF(reg_arg2,
							       index));
		}
		break;

	    case OP_CODE_VECTOR_SET:
		{
			int index, value;

			reg_arg2 = POP();
			reg_arg3 = POP();
			if (! CODE_VECTOR_P(reg_arg3) || ! FIXNUM_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			if (! FIXNUM_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			value = extract_fixnum(reg_val);

			index = extract_fixnum(reg_arg2);
			if (! VALID_INDEX_P(index,
					    CODE_VECTOR_LENGTH(reg_arg3)))
				raise_exception(EX_INDEX_OUT_OF_RANGE);
			CODE_VECTOR_SET(reg_arg3, index, value);
			reg_val = UNSPECIFIED_OBJ;
		}
		break;

	    case OP_STRING_P:
		reg_val = ENTER_BOOLEAN(STRING_P(reg_val));
		break;

	    case OP_MAKE_STRING:
		{
			int i, len, vec, value;

			reg_arg2 = POP();
			if (! FIXNUM_P(reg_arg2)
			    || (len = extract_fixnum(reg_arg2)) < 0)
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			if (! CHAR_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			ensure_space(STRING_SIZE(len));
			vec = MAKE_STRING(len);
			value = extract_char(reg_val);
			for (i = 0; i < len; i++)
				STRING_SET(vec, i, value);
			reg_val = vec;
		}
		break;

	    case OP_STRING_LENGTH:
		if (! STRING_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = enter_fixnum(STRING_LENGTH(reg_val));
		break;

	    case OP_STRING_REF:
		{
			int index;

			reg_arg2 = POP();
			if (! STRING_P(reg_arg2) || ! FIXNUM_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			index = extract_fixnum(reg_val);
			if (! VALID_INDEX_P(index, STRING_LENGTH(reg_arg2)))
				raise_exception(EX_INDEX_OUT_OF_RANGE);
			reg_val = ENTER_CHAR(STRING_REF(reg_arg2, index));
		}
		break;

	    case OP_STRING_SET:
		{
			int index, value;

			reg_arg2 = POP();
			reg_arg3 = POP();
			if (! STRING_P(reg_arg3) || ! FIXNUM_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			if (! CHAR_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			value = extract_char(reg_val);

			index = extract_fixnum(reg_arg2);
			if (! VALID_INDEX_P(index, STRING_LENGTH(reg_arg3)))
				raise_exception(EX_INDEX_OUT_OF_RANGE);
			STRING_SET(reg_arg3, index, value);
			reg_val = UNSPECIFIED_OBJ;
		}
		break;

	    case OP_VECTOR_P:
		reg_val = ENTER_BOOLEAN(VECTOR_P(reg_val));
		break;

	    case OP_MAKE_VECTOR:
		{
			int i, len, vec;

			reg_arg2 = POP();
			if (! FIXNUM_P(reg_arg2)
			    || (len = extract_fixnum(reg_arg2)) < 0)
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			ensure_space(VECTOR_SIZE(len));
			vec = MAKE_VECTOR(len);
			for (i = 0; i < len; i++)
				VECTOR_SET(vec, i, reg_val);
			reg_val = vec;
		}
		break;

	    case OP_VECTOR_LENGTH:
		if (! VECTOR_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = enter_fixnum(VECTOR_LENGTH(reg_val));
		break;

	    case OP_VECTOR_REF:
		{
			int index;

			reg_arg2 = POP();
			if (! VECTOR_P(reg_arg2) || ! FIXNUM_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			index = extract_fixnum(reg_val);
			if (! VALID_INDEX_P(index, VECTOR_LENGTH(reg_arg2)))
				raise_exception(EX_INDEX_OUT_OF_RANGE);
			reg_val = VECTOR_REF(reg_arg2, index);
		}
		break;

	    case OP_VECTOR_SET:
		{
			int index;

			reg_arg2 = POP();
			reg_arg3 = POP();
			if (! VECTOR_P(reg_arg3) || ! FIXNUM_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			index = extract_fixnum(reg_arg2);
			if (! VALID_INDEX_P(index, VECTOR_LENGTH(reg_arg3)))
				raise_exception(EX_INDEX_OUT_OF_RANGE);
			VECTOR_SET(reg_arg3, index, reg_val);
			reg_val = UNSPECIFIED_OBJ;
		}
		break;

		/*
		 *  Input/Output
		 */

	    case OP_OPEN_PORT:
		reg_arg2 = POP();
		if (! STRING_P(reg_arg2) || ! FIXNUM_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = open_port(reg_arg2, extract_fixnum(reg_val));
		break;

	    case OP_CLOSE_PORT:
		if (! PORT_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		close_port(reg_val);
		reg_val = UNSPECIFIED_OBJ;
		break;

	    case OP_INPUT_PORT_P:
		reg_val = ENTER_BOOLEAN(PORT_P(reg_val)
			&& extract_fixnum(PORT_MODE(reg_val)) == FOR_INPUT);
		break;

	    case OP_OUTPUT_PORT_P:
		reg_val = ENTER_BOOLEAN(PORT_P(reg_val)
			&& extract_fixnum(PORT_MODE(reg_val)) == FOR_OUTPUT);
		break;

	    case OP_READ_CHAR:
		if (! PORT_P(reg_val)
		    || extract_fixnum(PORT_MODE(reg_val)) != FOR_INPUT)
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = read_char(reg_val);
		break;

	    case OP_PEEK_CHAR:
		if (! PORT_P(reg_val)
		    || extract_fixnum(PORT_MODE(reg_val)) != FOR_INPUT)
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = peek_char(reg_val);
		break;

	    case OP_WRITE_CHAR:
		reg_arg2 = POP();
		if (! PORT_P(reg_val) || ! CHAR_P(reg_arg2))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		if (! OPEN_P(reg_val))
			raise_exception(EX_OPERATION_ON_CLOSED_PORT);

		putc(extract_char(reg_arg2), extract_port(reg_val));
		reg_val = UNSPECIFIED_OBJ;
		break;

	    case OP_WRITE_STRING:
		reg_arg2 = POP();
		if (! PORT_P(reg_val) || ! STRING_P(reg_arg2))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		if (! OPEN_P(reg_val))
			raise_exception(EX_OPERATION_ON_CLOSED_PORT);

		fprintf(extract_port(reg_val), "%s", extract_string(reg_arg2));
		reg_val = UNSPECIFIED_OBJ;
		break;


		/*
		 *  Miscellaneous stuff.
		 */

	    case OP_UNASSIGNED:
		reg_val = UNASSIGNED_MARKER_OBJ;
		break;

	    case OP_HALT:
		exit(extract_fixnum(reg_val));
		break;

	    case OP_SET_ENABLED_INTERRUPTS:
		{
			int temp;

			temp = reg_enabled_interrupts;
			reg_enabled_interrupts = extract_fixnum(reg_val);
			reg_val = enter_fixnum(temp);
		}
		break;

	    case OP_RETURN_FROM_HANDLER:
		reg_stack = reg_val;
		pop_istate();
		break;

	    case OP_WRITE_IMAGE:
		{
			char *str;

			if (! STRING_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			str = extract_string(reg_val);
			reg_val = UNSPECIFIED_OBJ;
			(void) push_istate();
			collect();
			write_heap(str);
			pop_istate();
			free(str);
		}
		break;

		/*
		 *  Unnecessary primitives
		 */

	    case OP_STRING_EQUAL_P:
		reg_arg2 = POP();
		if (! STRING_P(reg_arg2) || ! STRING_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = ENTER_BOOLEAN(STRING_EQUAL_P(reg_arg2, reg_val));
		break;

	    case OP_STRING_HASH:
		if (! STRING_P(reg_val))
			raise_exception(EX_WRONG_TYPE_ARGUMENT);
		reg_val = enter_fixnum(string_hash(reg_val));
		break;

	    case OP_REVERSE_LIST_TO_STRING:
		{
			int i, list, obj, length;

			reg_arg2 = POP();
			if (! FIXNUM_P(reg_val))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			if (! PAIR_P(reg_arg2) && reg_arg2 != NULL_OBJ)
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			list = reg_arg2;
			length = extract_fixnum(reg_val);
			ensure_space(STRING_SIZE(length));
			obj = MAKE_STRING(length);
			for (i = length - 1; i >= 0; i--)
			{
				STRING_SET(obj, i, extract_char(CAR(list)));
				list = CDR(list);
			}
			reg_val = obj;
		}
		break;

	    case OP_INTERN:
		reg_arg2 = POP();
		reg_val = intern(reg_arg2, reg_val);
		break;

	    case OP_LOOKUP:
		reg_arg2 = POP();
		reg_val = lookup(reg_val, reg_arg2);
		break;

		/*
		 *  New stuff.
		 */

	    case OP_RATIONAL_P:
		reg_val = ENTER_BOOLEAN(RATIONAL_P(reg_val));
		break;

	    case OP_REAL_P:
		reg_val = ENTER_BOOLEAN(REAL_P(reg_val));
		break;

	    case OP_COMPLEX_P:
		reg_val = ENTER_BOOLEAN(COMPLEX_P(reg_val));
		break;

	    case OP_DIVIDE:
		{
			int quotient, divisor, dividend;

			reg_arg2 = POP();
			if (! NUMBER_P(reg_val) || ! NUMBER_P(reg_arg2))
				raise_exception(EX_WRONG_TYPE_ARGUMENT);
			quotient = (dividend = extract_fixnum(reg_arg2))
				/ (divisor = extract_fixnum(reg_val));

			if (OVERFLOWS_P(quotient)
			    || quotient * divisor != dividend)
				raise_exception(EX_ARITHMETIC_OVERFLOW);
			reg_val = enter_fixnum(quotient);
		}
		break;

	    case OP_COLLECT:
		reg_val = UNSPECIFIED_OBJ;
		(void) push_istate();
		collect();
		pop_istate();
		break;

	    case OP_GET_DYNAMIC_STATE:
		reg_val = reg_dynamic_state;
		break;

	    case OP_SET_DYNAMIC_STATE:
		reg_dynamic_state = reg_val;
		reg_val = UNSPECIFIED_OBJ;
		break;

	    default:
		raise_exception(EX_UUO);
		break;
	}
	goto again;
}



/*
 *  The following functions are called infrequently, so they have been
 *  placed after the interpret function to guarantee that they will never
 *  be expanded inline by the compiler.
 */


void
handle_interrupt()
{
	if (! AVAILABLE_ON_STACK_P(ISTATE_SIZE))
		error("Out of memory while handling interrupt.");

	PREVIOUS_BYTE();
	PUSH(push_istate());
	reg_nargs = 1;
	reg_val = VECTOR_REF(GET_INTERRUPT_HANDLERS(), reg_pending_interrupt);
	reg_pending_interrupt = 0;
	reg_enabled_interrupts = 0;	/* disable interrupts */
	perform_application();

	/*
	 *  Jump back into the interpreter context and start executing
	 *  bytecodes again.
	 */
	longjmp(interpreter_context, 1);
}


void
raise_exception(exception)
int exception;
{
#ifdef DEBUG
	fprintf(stderr, "Exception is %d\n", exception);
#endif

	if (! AVAILABLE_ON_STACK_P(ISTATE_SIZE))
		error("Out of memory while handling exception.");
	PUSH(push_istate());
	reg_nargs = 1;
	reg_val = VECTOR_REF(GET_EXCEPTION_HANDLERS(), exception);
	perform_application();

	/*
	 *  Jump back into the interpreter context and start executing
	 *  bytecodes again.
	 */
	longjmp(interpreter_context, 1);
}
